import tokenize
import abc
from dataclasses import dataclass
from datetime import timedelta
from enum import IntEnum, auto
from io import StringIO
from re import match, search
from typing import Any

from cactus_test_definitions.errors import UnparseableVariableExpressionError

ConstantType = timedelta | int | float


@dataclass
class Token:
    """Custom token implementaion

    Attributes:
        string: representation of original token from input
        type: the kind of token found, an enum directly related from tokenize
        line: the input line that the token belongs
        start: coordinates of the token start wrt input
        end: coordinates of the token end wrt input
        param_key: optional to help with the special case of backfilling a self reference (i.e $this)
            to its underlying named value
    """

    string: str
    type: int
    line: str
    start: tuple[int, int]
    end: tuple[int, int]
    param_key: str | None = None

    @staticmethod
    def from_token_info(token_info: tokenize.TokenInfo, param_key: str | None = None) -> "Token":
        """Takes a tokenize.TokenInfo and returns an internal Token"""
        return Token(
            string=token_info.string,
            type=token_info.type,
            line=token_info.line,
            start=token_info.start,
            end=token_info.start,
            param_key=param_key,
        )


class NamedVariableType(IntEnum):
    # MUST resolve to a tz aware representation of the current datetime
    # Referenced in a test definition as $(now)
    NOW = auto()

    # MUST resolve to the "DERSetting.setMaxW" of the current EndDevice under test. Value in Watts
    # Referenced in a test definition as $(setMaxW)
    DERSETTING_SET_MAX_W = auto()
    DERSETTING_SET_MAX_VA = auto()
    DERSETTING_SET_MAX_VAR = auto()
    DERSETTING_SET_MAX_CHARGE_RATE_W = auto()
    DERSETTING_SET_MAX_DISCHARGE_RATE_W = auto()
    DERSETTING_SET_MAX_WH = auto()

    # Must resolve to DERCapablity of the current EndDevice under test
    DERCAPABILITY_RTG_MAX_VA = auto()  # VA ( after multiplier applied), reference $rtgMaxVA
    DERCAPABILITY_RTG_MAX_VAR = auto()  # VAr ( atfer multiplier applied), reference $rtgMaxVar
    DERCAPABILITY_RTG_MAX_W = auto()  # W ( after multiplier applied), reference $rtgMaxW
    DERCAPABILITY_RTG_MAX_CHARGE_RATE_W = auto()  # W ( after multiiplier applied), reference $rtgMaxChargeRateW
    DERCAPABILITY_RTG_MAX_DISCHARGE_RATE_W = auto()  # W ( after multiplier applied), reference $rtgMaxDischargeRateW
    DERCAPABILITY_RTG_MAX_WH = auto()  # Wh ( after multiplier applied), reference $rtgMaxWh

    # Storage extension
    DERSETTING_SET_MIN_WH = auto()
    DERCAPABILITY_NEG_RTG_MAX_CHARGE_RATE_W = auto()  # -W (after multiplier applied), reference $negRtgMaxChargeRateW


class OperationType(IntEnum):
    ADD = auto()
    SUBTRACT = auto()
    MULTIPLY = auto()
    DIVIDE = auto()
    EQ = auto()
    NE = auto()
    LT = auto()
    LTE = auto()
    GT = auto()
    GTE = auto()


OPERATION_MAPPINGS = {
    "+": OperationType.ADD,
    "-": OperationType.SUBTRACT,
    "*": OperationType.MULTIPLY,
    "/": OperationType.DIVIDE,
    "==": OperationType.EQ,
    "!=": OperationType.NE,
    "<": OperationType.LT,
    "<=": OperationType.LTE,
    ">": OperationType.GT,
    ">=": OperationType.GTE,
}


def snake_to_camel(snake: str) -> str:
    """Simple snake to camel case converter"""
    temp = snake.replace("_", " ").title().replace(" ", "")
    return temp[0].lower() + temp[1:]


def named_variable_repr(named_var: NamedVariableType) -> str:
    """Takes named variable enum and turns its name into its recognisable 2030.5 form"""
    name = named_var.name
    if len(name.split("_")) == 1:
        return snake_to_camel(name)
    match name.split("_", 1):
        case ["DERCAPABILITY", "NEG_RTG_MAX_CHARGE_RATE_W"]:
            return "(-DERCapability.rtgMaxChargeRateW)"
        case ["DERCAPABILITY", "RTG_MAX_VA"]:
            return "DERCapability.rtgMaxVA"
        case ["DERSETTING", "SET_MAX_VA"]:
            return "DERSetting.setMaxVA"
        case ["DERCAPABILITY", param_name]:
            return f"DERCapability.{snake_to_camel(param_name)}"
        case ["DERSETTING", param_name]:
            return f"DERSetting.{snake_to_camel(param_name)}"

    return snake_to_camel(name)


def operation_repr(op: OperationType) -> str:
    """Takes an operation type and returns its string representation"""
    operation_type_to_str_map: dict[OperationType, str] = {v: k for k, v in OPERATION_MAPPINGS.items()}
    return operation_type_to_str_map[op]


class BaseExpression(abc.ABC):
    """A base class for all expressions to inherit from"""

    @abc.abstractmethod
    def expression_representation(self) -> str:
        """Method for representing an expression human readably without overriding magic methods"""
        raise NotImplementedError


@dataclass
class Constant(BaseExpression):
    """Represents a constant value that doesn't require any test execution time resolution"""

    value: ConstantType  # The parsed value

    def expression_representation(self) -> str:
        return f"{self.value}"


@dataclass
class NamedVariable(BaseExpression):
    """A "NamedVariable" is value that can only be resolved at point during a test procedure execution (eg: as a
    Step's action is being applied). There are a fixed set of known variable types defined by NamedVariableType.

    Failures during resolving a variable (eg database doesn't have the data) MUST raise an exception
    """

    variable: NamedVariableType

    def expression_representation(self) -> str:
        return named_variable_repr(self.variable)


@dataclass
class Expression(BaseExpression):
    """An expression is a simple combination of two values that combine to make a single constant value. The operands
    can be constants or NamedVariables."""

    operation: OperationType
    lhs_operand: NamedVariable | Constant  # left hand side operand
    rhs_operand: NamedVariable | Constant  # right hand side operand

    def expression_representation(self) -> str:
        return " ".join(
            [
                f"{self.lhs_operand.expression_representation()}",
                f"{operation_repr(self.operation)}",
                f"{self.rhs_operand.expression_representation()}",
            ]
        )


def parse_time_delta(var_body: str) -> timedelta:
    """Parses a string like '5 minutes' into a representative timedelta"""

    m = match(r"(['\"])([0-9\-\.]*)\s*([^']*)(['\"])", var_body)
    if m is None:
        raise UnparseableVariableExpressionError(f"{var_body} can't be parsed into a timedelta")

    open_quote = m.group(1)
    number_string = m.group(2)
    time_unit_string = m.group(3).lower()
    close_quote = m.group(4)

    if open_quote != close_quote:
        raise UnparseableVariableExpressionError(f"{var_body} can't be parsed into a timedelta. Mismatching quotes")

    try:
        number = float(number_string)
    except ValueError:
        raise UnparseableVariableExpressionError(
            f"{var_body} can't be parsed into a timedelta. Bad number {number_string}"
        )

    if time_unit_string in {"day", "days"}:
        return timedelta(days=number)
    elif time_unit_string in {"hour", "hours", "hrs", "hr"}:
        return timedelta(hours=number)
    elif time_unit_string in {"minute", "minutes", "min", "mins"}:
        return timedelta(minutes=number)
    elif time_unit_string in {"second", "seconds", "sec", "secs"}:
        return timedelta(seconds=number)
    else:
        raise UnparseableVariableExpressionError(
            f"{var_body} can't be parsed into a timedelta. Unknown unit {time_unit_string}"
        )


def parse_unary_expression(token: Token) -> Constant | NamedVariable:
    """Parses a unary expression from a variable body"""

    if token.type == tokenize.NAME:
        # expect that a variable name is properly defined with correct case
        match token.string:
            case "now":
                return NamedVariable(NamedVariableType.NOW)
            case "this":
                if token.param_key == "this" or token.param_key is None:
                    raise UnparseableVariableExpressionError(f"$this cannot resolve to parameter {token.param_key}")
                # Modify token and maintain all other original data
                token.string = token.param_key
                token.param_key = None
                return parse_unary_expression(token)
            case "setMaxW":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)
            case "setMaxVA":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VA)
            case "setMaxVar":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VAR)
            case "setMaxChargeRateW":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_CHARGE_RATE_W)
            case "setMaxDischargeRateW":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_DISCHARGE_RATE_W)
            case "setMaxWh":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MAX_WH)
            case "rtgMaxVA":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_VA)
            case "rtgMaxVar":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_VAR)
            case "rtgMaxW":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_W)
            case "rtgMaxChargeRateW":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_CHARGE_RATE_W)
            case "rtgMaxDischargeRateW":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_DISCHARGE_RATE_W)
            case "rtgMaxWh":
                return NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_WH)
            # Storage extension
            case "setMinWh":
                return NamedVariable(NamedVariableType.DERSETTING_SET_MIN_WH)
            case "negRtgMaxChargeRateW":
                return NamedVariable(NamedVariableType.DERCAPABILITY_NEG_RTG_MAX_CHARGE_RATE_W)

        raise UnparseableVariableExpressionError(f"'{token.string}' isn't recognized as a named variable")

    try:
        if token.type == tokenize.NUMBER:
            if "." in token.string:
                return Constant(float(token.string))
            else:
                return Constant(int(token.string))
    except ValueError:
        raise UnparseableVariableExpressionError(f"'{token.string}' can't be converted to a number")

    if token.type == tokenize.STRING:
        return Constant(parse_time_delta(token.string))

    raise UnparseableVariableExpressionError(f"Unable to parse token {token}")


def parse_binary_expression(lhs_token: Token, operation: Token, rhs_token: Token) -> Expression:

    if operation.type != tokenize.OP:
        raise UnparseableVariableExpressionError(f"Expected an operation (eg + - / *) but found {operation}")

    operation_type = OPERATION_MAPPINGS.get(operation.string, None)
    if operation_type is None:
        raise UnparseableVariableExpressionError(f"Unable to parse operator {operation.string} into a OperationType")

    lhs = parse_unary_expression(lhs_token)
    rhs = parse_unary_expression(rhs_token)

    return Expression(operation=operation_type, lhs_operand=lhs, rhs_operand=rhs)


def parse_variable_expression_body(var_body: str, param_key: str | None) -> NamedVariable | Expression | Constant:
    """Given a variable definition: $(now - '5 seconds') - this function should be passed contents of that variable
    definition (the string within the parentheses) eg: "now - '5 seconds'

    Common variable patterns include,
    $(now) - Will return a tz aware datetime corresponding to the current moment in time
    $(now - '5 minute') - Same as above, but offset 5 minutes in the past
    $(0.5 * setMaxW) - 50% of the currently configured setMaxW for the current EndDevice

    Args:
        var_body: parseable expression
        param_key: the key that the expression body belongs e.g. setMaxW

    Returns:
        Parsed object

    Raises:
        UnparseableVariableExpressionError: on failed parsing attempt
    """
    if not var_body:
        raise UnparseableVariableExpressionError("var_body is empty/None")

    # Use the python parser to generate a simplified set of tokens representing the variable definition
    # Convert these into the internal representation of a Token
    try:
        var_tokens = [
            Token.from_token_info(t, param_key)
            for t in tokenize.generate_tokens(StringIO(var_body).readline)
            if t.type in {tokenize.NUMBER, tokenize.OP, tokenize.STRING, tokenize.NAME}
        ]
    except tokenize.TokenError as exc:
        raise UnparseableVariableExpressionError(f"Error tokenizing '{var_body}': {exc}")

    if len(var_tokens) == 1:
        return parse_unary_expression(var_tokens[0])
    elif len(var_tokens) == 3:
        return parse_binary_expression(var_tokens[0], var_tokens[1], var_tokens[2])
    else:
        raise UnparseableVariableExpressionError(f"Unable to parse {var_body} into a simple binary/unary expression")


def try_extract_variable_expression(body: Any) -> str | None:
    """Checks to see if a variable body (of any type) can be parsed by parse_variable_expression_body. If it can,
    it will be returned as a string. Otherwise None will be returned

    Can raise ValueError if body is a string appearing to contain a variable expression that is malformed (eg
    mismatching parentheses)"""
    if not isinstance(body, str):
        return None

    begin_variable_defn = body.find("$")
    if begin_variable_defn < 0:
        return None

    # The $ variable definition can be escaped with \$ so ensure it is checked
    if begin_variable_defn > 0 and (body[begin_variable_defn - 1] == "\\"):
        return None

    # At this point we are definitely parsing a variable expression. Failures here will raise ValueError
    if begin_variable_defn >= (len(body) - 1):
        raise ValueError(f"'{body}' appears to be a malformed variable definition. Try escaping $ like '\\$'")

    start_expr_body: int
    end_expr_body: int
    end_variable_defn: int  # First character after the end of the full variable definition
    if body[begin_variable_defn + 1] == "(":
        start_expr_body = begin_variable_defn + 2
        end_expr_body = body.index(")", start_expr_body)
        end_variable_defn = end_expr_body + 2
    else:
        start_expr_body = begin_variable_defn + 1

        remainder = body[start_expr_body:]
        match_variable = match(r"[_a-zA-Z0-9]*", remainder)
        if match_variable is None:
            raise ValueError(f"'{body}' appears to be a malformed variable definition")
        else:
            end_expr_body = start_expr_body + match_variable.end()
            end_variable_defn = end_expr_body + 1

    if start_expr_body == end_expr_body:
        raise ValueError(f"'{body}' appears to be a malformed variable definition.")
    variable_expression = body[start_expr_body:end_expr_body]

    # One last validation to look for leading/trailing text which would indicate
    leading_text = search(r"[^\s]", body[0:begin_variable_defn])
    if leading_text is not None:
        raise ValueError("Variable expressions must ONLY be a single variable with no other text")

    trailing_text = search(r"[^\s]", body[end_variable_defn:])
    if trailing_text is not None:
        raise ValueError("Variable expressions must ONLY be a single variable with no other text")

    return variable_expression


def is_resolvable_variable(v: Any) -> bool:
    """Returns True if the supplied value is a variable definition that requires resolving"""
    return isinstance(v, NamedVariable) or isinstance(v, Expression) or isinstance(v, Constant)


def has_named_variable(
    parameter_value: NamedVariable | Expression | Constant, named_variable: NamedVariableType
) -> bool:
    """Return True if the supplied named variable is used in the the parameter 'parameter_value'"""

    if isinstance(parameter_value, Constant):
        return False

    if isinstance(parameter_value, NamedVariable):
        return parameter_value.variable == named_variable

    if isinstance(parameter_value, Expression):
        if isinstance(parameter_value.lhs_operand, NamedVariable):
            if parameter_value.lhs_operand.variable == named_variable:
                return True
        if isinstance(parameter_value.rhs_operand, NamedVariable):
            if parameter_value.rhs_operand.variable == named_variable:
                return True
        return False



from datetime import datetime

import pytest
from cactus_test_definitions import variable_expressions as varexps
from cactus_test_definitions.client.actions import (
    Action,
    validate_action_parameters,
)
from cactus_test_definitions.errors import TestProcedureDefinitionError


@pytest.mark.parametrize(
    "action, is_valid",
    [
        (Action("foo", {}), False),
        (Action("enable-steps", {"steps": []}), True),
        (Action("enable-steps", {"steps": ["other-step"]}), True),
        (Action("enable-steps", {"steps": ["other-step"], "unused-param": 123}), False),  # extra param
        (Action("set-default-der-control", {}), True),
        (Action("set-default-der-control", {"opModImpLimW": 12.3, "opModExpLimW": 14}), True),
        # storage extension specific
        (
            Action("set-default-der-control", {"opModImpLimW": 12.3, "opModExpLimW": 14, "opModStorageTargetW": 200}),
            True,
        ),
        (Action("set-default-der-control", {"opModImpLimW": 12.3, "opModExpLimW": "Invalid"}), False),  # bad type
        (Action("create-der-control", {"start": datetime(2022, 1, 3)}), False),  # Missing duration_seconds
        (Action("create-der-control", {"start": datetime(2022, 1, 3), "duration_seconds": 123}), True),
        # storage extension specific
        (
            Action(
                "create-der-control",
                {"start": datetime(2022, 1, 3), "duration_seconds": 123, "opModStorageTargetW": 1.5},
            ),
            True,
        ),
    ],
)
def test_validate_action_parameters(action: Action, is_valid: bool):
    if is_valid:
        validate_action_parameters("foo", "bar", action)
    else:
        with pytest.raises(TestProcedureDefinitionError):
            validate_action_parameters("foo", "bar", action)


def test_action_expression() -> None:
    """Tests the creation of an Action that has an expression as one of its parameters"""
    type_str = "some_action"
    params = {"setMaxW": "$(this == rtgMaxW)"}
    action = Action(type_str, params)

    check_set_max_w = action.parameters["setMaxW"]
    assert isinstance(check_set_max_w, varexps.Expression)
    assert check_set_max_w.operation == varexps.OperationType.EQ
    assert check_set_max_w.lhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERSETTING_SET_MAX_W)
    assert check_set_max_w.rhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERCAPABILITY_RTG_MAX_W)


def test_action_expression_using_storage_extension() -> None:
    """Tests the creation of an Action that has an expression as one of its parameters.

    Using storage extension params."""
    type_str = "some_action"
    params = {"setMinWh": "$(this <= setMaxWh)"}
    action = Action(type_str, params)

    check_set_min_wh = action.parameters["setMinWh"]
    assert isinstance(check_set_min_wh, varexps.Expression)
    assert check_set_min_wh.operation == varexps.OperationType.LTE
    assert check_set_min_wh.lhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERSETTING_SET_MIN_WH)
    assert check_set_min_wh.rhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERSETTING_SET_MAX_WH)



from cactus_test_definitions.csipaus import CSIPAusVersion
from cactus_test_definitions.errors import (
    TestProcedureDefinitionError,
    UnparseableVariableExpressionError,
    UnresolvableVariableError,
)
from cactus_test_definitions.variable_expressions import (
    Constant,
    ConstantType,
    Expression,
    NamedVariable,
    NamedVariableType,
    OperationType,
    parse_binary_expression,
    parse_time_delta,
    parse_unary_expression,
    parse_variable_expression_body,
    try_extract_variable_expression,
)

__version__ = "1.7.1"

__all__ = [
    "TestProcedureDefinitionError",
    "CSIPAusVersion",
    "Expression",
    "Constant",
    "ConstantType",
    "NamedVariable",
    "OperationType",
    "UnresolvableVariableError",
    "UnparseableVariableExpressionError",
    "NamedVariableType",
    "try_extract_variable_expression",
    "parse_variable_expression_body",
    "parse_time_delta",
    "parse_binary_expression",
    "parse_unary_expression",
]



import pytest
from cactus_test_definitions.server.actions import ACTION_PARAMETER_SCHEMA
from cactus_test_definitions.server.test_procedures import (
    TestProcedureId,
    get_test_procedure,
)
from cactus_test_definitions.server.validate import validate_test_procedure


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_TestProcedure_individually_valid(tp_id: TestProcedureId):
    """Tests that each TestProcedureId can be loaded via get_test_procedure without issue AND that it
    passes validate_test_procedure."""
    tp = get_test_procedure(tp_id)
    assert tp.steps, "There should be steps configured"
    validate_test_procedure(tp, tp_id)


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_each_step_id_unique(tp_id: TestProcedureId):
    tp = get_test_procedure(tp_id)
    all_ids = [s.id for s in tp.steps]
    assert list(sorted(all_ids)) == list(sorted(set(all_ids)))
    assert len(set((s.id for s in tp.steps))) == len(tp.steps), "All steps must have a unique id property"


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_each_alias_defined(tp_id: str):
    """Ensures that each test procedure's steps that have actions using an alias... define those aliases in advance"""

    tp = get_test_procedure(tp_id)

    # sanity check - make sure we are looking for the action names that are valid
    UPSERT_MUP_ACTION = "upsert-mup"
    INSERT_READING_ACTION = "insert-readings"
    CREATE_SUB_ACTION = "create-subscription"
    DELETE_SUB_ACTION = "delete-subscription"
    assert UPSERT_MUP_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"
    assert (
        INSERT_READING_ACTION in ACTION_PARAMETER_SCHEMA
    ), "If this fails - the action name has changed. Update this test"
    assert CREATE_SUB_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"
    assert DELETE_SUB_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"

    mup_aliases_found: set[str] = set()
    sub_aliases_found: set[str] = set()
    for step in tp.steps:
        if step.action.parameters:
            mup_id = step.action.parameters.get("mup_id", None)
            sub_id = step.action.parameters.get("sub_id", None)
        else:
            mup_id = None
            sub_id = None

        if step.action.type == UPSERT_MUP_ACTION:
            mup_aliases_found.add(mup_id)
        elif step.action.type == INSERT_READING_ACTION:
            assert mup_id and (mup_id in mup_aliases_found), "mup_id hasn't been defined yet."
        elif step.action.type == CREATE_SUB_ACTION:
            assert sub_id and (sub_id not in sub_aliases_found), "sub_id is already defined in this test."
            sub_aliases_found.add(sub_id)
        elif step.action.type == DELETE_SUB_ACTION:
            assert sub_id and (sub_id in sub_aliases_found), "sub_id hasn't been defined yet."

        # Any instance of "sub_id" / "mup_id" in a check must have ALSO been defined in an earlier action
        for check in step.checks or []:
            sub_id = check.parameters.get("sub_id", None)
            if sub_id:
                # This will match
                assert sub_id and (sub_id in sub_aliases_found), "sub_id hasn't been defined yet."

            mup_id = check.parameters.get("mup_id", None)
            if mup_id:
                # This will match
                assert mup_id and (mup_id in mup_aliases_found), "mup_id hasn't been defined yet."



from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any

import pytest
from cactus_test_definitions.client.test_procedures import Action
from cactus_test_definitions.variable_expressions import (
    Constant,
    Expression,
    NamedVariable,
    NamedVariableType,
    OperationType,
    UnparseableVariableExpressionError,
    has_named_variable,
    is_resolvable_variable,
    parse_time_delta,
    parse_variable_expression_body,
    try_extract_variable_expression,
    snake_to_camel,
    named_variable_repr,
    operation_repr,
    BaseExpression,
)


@pytest.mark.parametrize(
    "body, expected",
    [
        (None, None),
        (123, None),
        (12.3, None),
        (["$no_list_inspection"], None),
        ({"$no_dict_inspection": "$no_dict_inspection"}, None),
        (datetime(2001, 2, 3), None),
        ("", None),
        ("  ", None),
        ("$now", "now"),
        ("  $now\n  ", "now"),
        ("$(now)", "now"),
        ("$(now - '5 minutes')", "now - '5 minutes'"),
        ("$(  now - '5 minutes'  )", "  now - '5 minutes'  "),
        ("\\$(now)", None),  # escaped
        ("  \\$(now) ", None),  # escaped
        ("  \\$now ", None),  # escaped
        ("\\$now", None),  # escaped
        ("$variable_with_556", "variable_with_556"),
        ("$(variable_with_556)", "variable_with_556"),
        ("$variable-556", ValueError),  # Don't allow "non variable" data to get omitted
        ("$(variable-556)", "variable-556"),
        ("$longer_variable no spaces", ValueError),
        ("$(longer_variable but include everything)", "longer_variable but include everything"),
        ("$(now", ValueError),  # Unclosed bracket
        (" $(now  ", ValueError),  # Unclosed bracket
        (" $(now foo", ValueError),  # Unclosed bracket
        ("$ invalid_space", ValueError),
        ("$-invalid_char", ValueError),
        ("$", ValueError),  # No variable body included
        ("$ ", ValueError),  # No variable body included
        ("\\$", None),
        ("$()", ValueError),  # No variable body included
        ("\\$()", None),
        (" \\$() ", None),
        ("$(valid) no trailing data", ValueError),  # A string is either a variable or not. Can't substring variables
        ("no leading data $(valid)", ValueError),  # A string is either a variable or not. Can't substring variables
    ],
)
def test_try_extract_variable_expression(body: Any, expected: str | None | type[Exception]):
    """Various test cases that try expose edge cases in try_extract_variable_expression"""
    if isinstance(expected, type):
        with pytest.raises(expected):
            try_extract_variable_expression(body)
    else:
        actual = try_extract_variable_expression(body)
        assert actual is None or isinstance(actual, str)
        assert actual == expected, f"Input string: '{body}'"


@pytest.mark.parametrize(
    "unquoted_raw, expected",
    [
        ("", UnparseableVariableExpressionError),
        ("5 mins", timedelta(minutes=5)),
        ("-5 mins", timedelta(minutes=-5)),
        ("5.23 mins", timedelta(minutes=5.23)),
        ("-5.23 mins", timedelta(minutes=-5.23)),
        ("-0.03 days", timedelta(days=-0.03)),
        ("-0.03   days", timedelta(days=-0.03)),
        ("-0.03days", timedelta(days=-0.03)),
        ("4 minutes", timedelta(minutes=4)),
        ("4 minute", timedelta(minutes=4)),
        ("0 minute", timedelta(minutes=0)),
        ("4 hours", timedelta(hours=4)),
        ("4 hour", timedelta(hours=4)),
        ("4 hrs", timedelta(hours=4)),
        ("4 hr", timedelta(hours=4)),
        ("4 days", timedelta(days=4)),
        ("4 day", timedelta(days=4)),
        ("4 seconds", timedelta(seconds=4)),
        ("4 second", timedelta(seconds=4)),
        ("4 secs", timedelta(seconds=4)),
        ("4 sec", timedelta(seconds=4)),
        ("4 foo", UnparseableVariableExpressionError),  # Unknown unit
        ("4-.2 foo", UnparseableVariableExpressionError),  # Not a number
        ("4.2.3 foo", UnparseableVariableExpressionError),  # Not a number
    ],
)
def test_parse_time_delta(unquoted_raw: str, expected: timedelta | type[Exception]):
    """Tests the various ways a time delta interval can be encoded (and the various ways it can fail)"""
    VALID_QUOTE_CHARS = ["'", '"']

    # Check weird edge cases from unquoted intervals or badly quoted intervals
    with pytest.raises(UnparseableVariableExpressionError):
        parse_time_delta(unquoted_raw)

    for quote_char in ["'", '"']:
        with pytest.raises(UnparseableVariableExpressionError):
            parse_time_delta(quote_char + unquoted_raw)
        with pytest.raises(UnparseableVariableExpressionError):
            parse_time_delta(unquoted_raw + quote_char)

    with pytest.raises(UnparseableVariableExpressionError):
        parse_time_delta(VALID_QUOTE_CHARS[0] + quote_char + VALID_QUOTE_CHARS[1])
    with pytest.raises(UnparseableVariableExpressionError):
        parse_time_delta(VALID_QUOTE_CHARS[1] + quote_char + VALID_QUOTE_CHARS[0])

    # Now do the actual test
    if isinstance(expected, type):
        for quote_char in VALID_QUOTE_CHARS:
            with pytest.raises(expected):
                parse_time_delta(quote_char + unquoted_raw + quote_char)
    else:
        for quote_char in VALID_QUOTE_CHARS:
            actual = parse_time_delta(quote_char + unquoted_raw + quote_char)
            assert isinstance(actual, timedelta)
            assert actual == expected, f"Input string: {quote_char + unquoted_raw + quote_char}"


@pytest.mark.parametrize(
    "var_body, expected",
    [
        ("1.23", Constant(1.23)),
        ("123", Constant(123)),
        ("  123  ", Constant(123)),
        ("'-4.56 hours'", Constant(timedelta(hours=-4.56))),
        ('"0.12 days"', Constant(timedelta(days=0.12))),
        (" \t  '-4.56 hours' \t  ", Constant(timedelta(hours=-4.56))),
        ("now", NamedVariable(NamedVariableType.NOW)),
        ("NOW", UnparseableVariableExpressionError),  # case sensitive
        ("setMaxW", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)),
        ("SETMAXW", UnparseableVariableExpressionError),  # case sensitive
        ("foo", UnparseableVariableExpressionError),  # unknown named variable
        (
            "0.5 * 0.2",
            Expression(OperationType.MULTIPLY, Constant(0.5), Constant(0.2)),
        ),
        (
            "0.5 * setMaxW",
            Expression(OperationType.MULTIPLY, Constant(0.5), NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)),
        ),
        (
            "setMaxW / 2",
            Expression(OperationType.DIVIDE, NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W), Constant(2)),
        ),
        (
            "  now  -    '-12   minutes'  ",
            Expression(OperationType.SUBTRACT, NamedVariable(NamedVariableType.NOW), Constant(timedelta(minutes=-12))),
        ),
        (
            "now-'-12minutes'",
            Expression(OperationType.SUBTRACT, NamedVariable(NamedVariableType.NOW), Constant(timedelta(minutes=-12))),
        ),
        (
            'now + "3 day"',
            Expression(OperationType.ADD, NamedVariable(NamedVariableType.NOW), Constant(timedelta(days=3))),
        ),
        (
            '"3 day" < "5 day"',
            Expression(OperationType.LT, Constant(timedelta(days=3)), Constant(timedelta(days=5))),
        ),
        (
            "rtgMaxW / 2",
            Expression(OperationType.DIVIDE, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_W), Constant(2)),
        ),
        ("setMaxVA", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VA)),
        ("setMaxVar", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VAR)),
        ("setMaxChargeRateW", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_CHARGE_RATE_W)),
        ("setMaxDischargeRateW", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_DISCHARGE_RATE_W)),
        ("setMaxWh", NamedVariable(NamedVariableType.DERSETTING_SET_MAX_WH)),
        (
            "rtgMaxVar == 5.0",
            Expression(OperationType.EQ, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_VAR), Constant(5.0)),
        ),
        (
            "rtgMaxW != 0.5",
            Expression(OperationType.NE, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_W), Constant(0.5)),
        ),
        (
            "rtgMaxChargeRateW <= 0.5",
            Expression(
                OperationType.LTE, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_CHARGE_RATE_W), Constant(0.5)
            ),
        ),
        (
            "rtgMaxDischargeRateW > 0.5",
            Expression(
                OperationType.GT, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_DISCHARGE_RATE_W), Constant(0.5)
            ),
        ),
        (
            "rtgMaxWh >= 0.5",
            Expression(OperationType.GTE, NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_WH), Constant(0.5)),
        ),
        ("now + foo", UnparseableVariableExpressionError),
        ("now foo +", UnparseableVariableExpressionError),
        ("now foo ", UnparseableVariableExpressionError),
        ("7 + + 8 ", UnparseableVariableExpressionError),
        ("'5 mins + now ", UnparseableVariableExpressionError),  # Unterminated string literal
        ("now + '5 mins", UnparseableVariableExpressionError),  # Unterminated string literal
        # Storage extension
        (
            "setMinWh >= 0.5",
            Expression(OperationType.GTE, NamedVariable(NamedVariableType.DERSETTING_SET_MIN_WH), Constant(0.5)),
        ),
        (
            "negRtgMaxChargeRateW + 50",
            Expression(
                OperationType.ADD,
                NamedVariable(NamedVariableType.DERCAPABILITY_NEG_RTG_MAX_CHARGE_RATE_W),
                Constant(50),
            ),
        ),
    ],
)
def test_parse_variable_expression_body(
    var_body: str, expected: type[Exception] | NamedVariable | Constant | Expression
):
    """Top level parsing test to ensure that a variety of variable bodies parse (or fail) in an expected fashion"""

    if isinstance(expected, type):
        with pytest.raises(expected):
            parse_variable_expression_body(var_body, None)
    else:
        actual = parse_variable_expression_body(var_body, None)
        assert isinstance(actual, NamedVariable) or isinstance(actual, Constant) or isinstance(actual, Expression)
        assert actual == expected, f"Input string: {var_body}"


@pytest.mark.parametrize(
    "var_body,expected,param_key",
    [
        (
            "this < rtgMaxVA",
            Expression(
                OperationType.LT,
                NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VA),
                NamedVariable(NamedVariableType.DERCAPABILITY_RTG_MAX_VA),
            ),
            "setMaxVA",
        ),
        (
            "this < rtgMaxVA",
            UnparseableVariableExpressionError,
            "someUndefinedNamedVariable",
        ),
    ],
)
def test_parse_variable_expression_body_this(
    var_body: str, expected: type[Exception] | NamedVariable | Constant | Expression, param_key: str | None
) -> None:
    """Parsing to ensure that variable bodies that contain `this` parse (or fail) as expected"""
    if isinstance(expected, type):
        with pytest.raises(expected):
            parse_variable_expression_body(var_body, param_key)
    else:
        actual = parse_variable_expression_body(var_body, param_key)
        assert isinstance(actual, NamedVariable) or isinstance(actual, Constant) or isinstance(actual, Expression)
        assert actual == expected, f"Input string: {var_body}"


@pytest.mark.parametrize(
    "input, expected",
    [
        (None, False),
        ("", False),
        ("string value", False),
        (123, False),
        (1.23, False),
        (Decimal("1.2"), False),
        (datetime(2022, 11, 3), False),
        (timedelta(2), False),
        (Action("", {}), False),
        ([], False),
        ({}, False),
        (NamedVariable(NamedVariableType.NOW), True),
        (NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W), True),
        (Constant(1.23), True),
        (Constant(timedelta(5)), True),
        (Expression(OperationType.ADD, Constant(1.23), NamedVariable(NamedVariableType.NOW)), True),
    ],
)
def test_is_resolvable_variable(input: Any, expected: bool):
    result = is_resolvable_variable(input)
    assert isinstance(result, bool)
    assert result == expected


@pytest.mark.parametrize(
    "parameter,variable,expected",
    [
        (Constant(value=0), NamedVariableType.DERSETTING_SET_MAX_W, False),
        (Constant(value=5), NamedVariableType.DERSETTING_SET_MAX_W, False),
        (NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W), NamedVariableType.DERSETTING_SET_MAX_W, True),
        (NamedVariable(NamedVariableType.DERSETTING_SET_MAX_VA), NamedVariableType.DERSETTING_SET_MAX_W, False),
        (
            Expression(OperationType.ADD, Constant(1.23), NamedVariable(NamedVariableType.NOW)),
            NamedVariableType.DERSETTING_SET_MAX_W,
            False,
        ),
        (
            Expression(OperationType.ADD, NamedVariable(NamedVariableType.NOW), Constant(1.23)),
            NamedVariableType.DERSETTING_SET_MAX_W,
            False,
        ),
        (
            Expression(OperationType.ADD, Constant(1.23), NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)),
            NamedVariableType.DERSETTING_SET_MAX_W,
            True,
        ),
        (
            Expression(OperationType.ADD, NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W), Constant(1.23)),
            NamedVariableType.DERSETTING_SET_MAX_W,
            True,
        ),
    ],
)
def test_has_named_variable(
    parameter: NamedVariable | Expression | Constant, variable: NamedVariableType, expected: bool
):
    result = has_named_variable(parameter_value=parameter, named_variable=variable)
    assert isinstance(result, bool)
    assert result == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        ("SET_MAX_W", "setMaxW"),
        ("SET_MAX_VAR", "setMaxVar"),
        ("NOW", "now"),
    ],
)
def test_snake_to_camel(input: str, expected: str) -> None:
    assert snake_to_camel(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (NamedVariableType.DERCAPABILITY_RTG_MAX_VAR, "DERCapability.rtgMaxVar"),
        (NamedVariableType.DERSETTING_SET_MIN_WH, "DERSetting.setMinWh"),
        (NamedVariableType.DERCAPABILITY_RTG_MAX_VA, "DERCapability.rtgMaxVA"),
        (NamedVariableType.DERSETTING_SET_MAX_VA, "DERSetting.setMaxVA"),
        (NamedVariableType.DERCAPABILITY_NEG_RTG_MAX_CHARGE_RATE_W, "(-DERCapability.rtgMaxChargeRateW)"),
        (NamedVariableType.NOW, "now"),
    ],
)
def test_named_variable_repr(input: NamedVariableType, expected: str) -> None:
    assert named_variable_repr(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (OperationType.ADD, "+"),
        (OperationType.MULTIPLY, "*"),
        (OperationType.SUBTRACT, "-"),
        (OperationType.DIVIDE, "/"),
        (OperationType.EQ, "=="),
        (OperationType.GT, ">"),
        (OperationType.GTE, ">="),
        (OperationType.LT, "<"),
        (OperationType.LTE, "<="),
        (OperationType.NE, "!="),
    ],
)
def test_operation_repr(input: OperationType, expected: str) -> None:
    assert operation_repr(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (NamedVariable(variable=NamedVariableType.DERCAPABILITY_RTG_MAX_VA), "DERCapability.rtgMaxVA"),
        (
            Expression(OperationType.ADD, NamedVariable(NamedVariableType.DERSETTING_SET_MIN_WH), Constant(5.5)),
            "DERSetting.setMinWh + 5.5",
        ),
        (Constant(654.456), "654.456"),
    ],
)
def test_base_expression_expression_representation(input: BaseExpression, expected: str) -> None:
    assert input.expression_representation() == expected



import pytest
from cactus_test_definitions import variable_expressions as varexps
from cactus_test_definitions.client.checks import Check, validate_check_parameters
from cactus_test_definitions.errors import TestProcedureDefinitionError


@pytest.mark.parametrize(
    "check, is_valid",
    [
        (Check("foo", {}), False),  # Unknown check
        (Check("readings-site-active-power", {}), True),
        (Check("readings-site-active-power", {"minimum_count": "3"}), False),
        (Check("readings-site-active-power", {"minimum_count": 3}), True),
        (Check("readings-site-active-power", {"minimum_count": 3, "minimum_level": 12.3}), True),
        (Check("readings-site-active-power", {"minimum_count": 3, "maximum_level": 12.3}), True),
        (Check("readings-site-active-power", {"minimum_count": 3, "minimum_level": 12.3, "maximum_level": 12.3}), True),
        (
            Check(
                "readings-site-active-power",
                {"minimum_count": 3, "minimum_level": 12.3, "maximum_level": 12.3, "window_seconds": 12.3},
            ),
            False,
        ),
        (
            Check(
                "readings-site-active-power",
                {"minimum_count": 3, "minimum_level": 12.3, "maximum_level": 12.3, "window_seconds": 180},
            ),
            True,
        ),
        (
            Check(
                "readings-site-active-power",
                {"minimum_count": 3, "minimum_level": 12.3, "maximum_level": 12.3, "window_seconds": -180},
            ),
            False,
        ),
        (Check("der-settings-contents", {"setGradW": 27}), True),
        (Check("der-settings-contents", {"doeModesEnabled_set": "0f"}), True),
        (Check("der-settings-contents", {"doeModesEnabled_unset": "0f"}), True),
        (Check("der-settings-contents", {"modesEnabled_set": "E"}), True),
        (Check("der-settings-contents", {"modesEnabled_unset": "E"}), True),
        (Check("der-settings-contents", {"doeModesEnabled_set": 12}), False),
        (Check("der-settings-contents", {"doeModesEnabled_unset": 12}), False),
        # Storage extension
        (Check("der-settings-contents", {"vppModesEnabled_set": "1"}), True),
        (Check("der-settings-contents", {"vppModesEnabled_unset": "1"}), True),
        (Check("der-capability-contents", {"vppModesSupported_set": "1"}), True),
        (Check("der-capability-contents", {"vppModesSupported_unset": "1"}), True),
        (Check("readings-der-stored-energy", {"minimum_count": 3}), True),
    ],
)
def test_validate_check_parameters(check: Check, is_valid: bool):
    if is_valid:
        validate_check_parameters("foo", check)
    else:
        with pytest.raises(TestProcedureDefinitionError):
            validate_check_parameters("foo", check)


def test_check_expression() -> None:
    """Tests the creation of a Check that has an expression as one of its parameters"""
    type_str = "some_check"
    params = {"setMaxW": "$(this < rtgMaxW)"}
    check = Check(type_str, params)

    check_set_max_w = check.parameters["setMaxW"]
    assert isinstance(check_set_max_w, varexps.Expression)
    assert check_set_max_w.operation == varexps.OperationType.LT
    assert check_set_max_w.lhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERSETTING_SET_MAX_W)
    assert check_set_max_w.rhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERCAPABILITY_RTG_MAX_W)
